const express = require('express')
const bearerToken = require('express-bearer-token')
const bodyParser = require('body-parser')
const _ = require('lodash')
const app = express()
const port = process.env.PORT || 8080

const VALID_TOKEN = 'patt3rnl0ckk3y'

// The valid moves from each node
const VALID_MOVES = {
  1: ['2', '4', '5'],
  2: ['1', '3', '4', '5', '6'],
  3: ['2', '5', '6'],
  4: ['1', '2', '5', '7', '8'],
  5: ['1', '2', '3', '4', '6', '7', '8', '9'],
  6: ['2', '3', '5', '8', '9'],
  7: ['4', '5', '8'],
  8: ['4', '5', '6', '7', '9'],
  9: ['5', '6', '8']
}

//  the x, y coordinates for the numbers in the canvas
const startingPoints = {}

app.use(bearerToken())
app.use(bodyParser.json())

app.use((req, res, next) => {
  if (req.token === VALID_TOKEN) {
    next()
  } else {
    res.status(401).send({ message: 'Invalid key' })
  }
})

app.post('/pattern-lock', (req, res) => {
  const number = req.body.number

  const isValid = isNumberValid(number)

  if (!isValid) {
    res.status(400).send({ message: 'Invalid number' })

    return
  }

  let validPatterns = []

  // First check if the provided number pattern is a loop
  // It will begin and end with same number
  let isLoopPattern = false
  const numberAsString = number.toString()

  if (numberAsString[0] === numberAsString[numberAsString.length - 1]) {
    isLoopPattern = true
  }

  if (!isLoopPattern) {
    // The possible patterns are the number and its reverse
    validPatterns = [number, parseInt(numberAsString.split('').reverse().join(''), 10)]
  } else {
    validPatterns = [number, parseInt(numberAsString.split('').reverse().join(''), 10)]

    let newNumber = numberAsString

    for (let i = 0; i < numberAsString.length - 2; i++) {
      newNumber = getNextNewNumber(newNumber)
      validPatterns.push(parseInt(newNumber, 10))
      validPatterns.push(parseInt(newNumber.split('').reverse().join(''), 10))
    }
  }

  validPatterns = _.uniq(validPatterns)

  res.send({ 'possible-numbers': validPatterns })
})

app.listen(port, () => {
  console.log(`App listening on port ${port}`)
})

/**
 * Checks if the number is a valid pattern
 * @param {Number} number The number to check
 */
function isNumberValid (number) {
  if (!_.isNumber(number)) {
    return false
  }

  const numberAsString = number.toString()
  let isValid = true
  const uniqueNumbers = []

  if (numberAsString.length < 2 || numberAsString.length > 5) {
    isValid = false
  } else {
    for (let i = 0; i < numberAsString.length - 1; i++) {
      const currentDigit = numberAsString[i]
      const nextDigit = numberAsString[i + 1]

      if (!VALID_MOVES[currentDigit].includes(nextDigit)) {
        isValid = false
        break
      } else if (uniqueNumbers.includes(currentDigit)) {
        // A number cannot be present twice in the pattern, unless it is the first and last digit in the number
        isValid = false
        break
      }

      uniqueNumbers.push(currentDigit)
    }

    const lastDigit = numberAsString[numberAsString.length - 1]

    if (uniqueNumbers.includes(lastDigit) && numberAsString[0] !== lastDigit) {
      // A number cannot be present twice in the pattern, unless it is the first and last digit in the number
      isValid = false
    }
  }

  return isValid
}

/**
 * Moves the first digit to the last and returns the number as string
 * @param {String} numberAsString The number as a string
 */
function getNextNewNumber (numberAsString) {
  numberAsString = numberAsString.substring(1)
  const firstChar = numberAsString[0]
  numberAsString = `${numberAsString}${firstChar}`

  return numberAsString
}
